from ..parser import (
    bbcode_from_text,
    parse_user_input,
    get_valid_bbcode_packs,
    deserialize,
    BBCodeParseError,
)

import unittest
import re


class TestBBCodeParser(unittest.TestCase):
    def test_get_packs(self):
        result = get_valid_bbcode_packs()
        self.assertIn("message", result)
        self.assertIn("web_page", result)

    def test_invalid_pack(self):
        with self.assertRaisesRegex(ValueError, r"Unknown BBCode pack"):
            parse_user_input("blabla", "")

    def test_serialize_deserialize(self):
        node = parse_user_input("message", "[b]Test é ><[/b]")
        self.assertEqual(node.serialize(), "\x01[b]Test é ><[/b]")

        node = deserialize("message", "\x01[b]Test é ><[/b]")
        self.chk_bbcode(node, "[b]Test é ><[/b]")

        with self.chk_err("Version 10 incorrecte"):
            deserialize("message", "\x0A[b]Test é ><[/b]")

    def test_from_text(self):
        node = bbcode_from_text("[b]Test é ><")
        self.chk_bbcode(node, "[b]Test é ><")
        self.chk_html(node, "[b]Test é &gt;&lt;")
        self.chk_text(node, "[b]Test é ><")

    def test_text_node(self):
        node = parse_user_input("message", "Test é ><")
        self.chk_bbcode(node, "Test é ><")
        self.chk_html(node, "Test é &gt;&lt;")
        self.chk_text(node, "Test é ><")

        node = parse_user_input("message", "Test é ><")
        self.chk_bbcode(node, "Test é ><")
        self.chk_html(node, "Test é &gt;&lt;")
        self.chk_text(node, "Test é ><")

    def test_simple_text(self):
        node = parse_user_input("message", "Hello, world ééé\\\"'`<>“")
        self.chk_bbcode(node, "Hello, world ééé\\\"'`<>“")
        self.chk_html(node, "Hello, world ééé\\&quot;&#x27;`&lt;&gt;“")
        self.chk_text(node, "Hello, world ééé\\\"'`<>“")

    def test_return_text(self):
        node = parse_user_input("message", "a\nb")
        self.chk_bbcode(node, "a\nb")
        self.chk_html(node, "a<br />\nb")
        self.chk_text(node, "a\nb")

        node = parse_user_input("message", "a\n")
        self.chk_bbcode(node, "a\n")
        self.chk_html(node, "a\n")
        self.chk_text(node, "a\n")

    def test_br(self):
        node = parse_user_input("message", "[b]a[br][/b]")
        self.chk_bbcode(node, "[b]a[br][/b]")
        self.chk_html(node, "<b>a<br /></b>")
        self.chk_text(node, "a")

    def test_br_close(self):
        node = parse_user_input("message", "[b]a[br]x[/br][/b]")
        self.chk_bbcode(node, "[b]a[br]x[/br][/b]")
        self.chk_html(node, "<b>a<br />x[/br]</b>")
        self.chk_text(node, "ax[/br]")

    def test_correct_nesting(self):
        node = parse_user_input("message", "[b]g [i]i[/i] [u]u[/u]...[/b]")
        self.chk_bbcode(node, "[b]g [i]i[/i] [u]u[/u]...[/b]")
        self.chk_html(
            node,
            "<b>g <i>i</i> " '<span style="text-decoration:underline">u</span>...</b>',
        )
        self.chk_text(node, "g i u...")

    def test_no_closing(self):
        with self.chk_err("BBCode [b] mal fermé."):
            parse_user_input("message", "[b]g")

        with self.chk_err("BBCode [code] mal fermé."):
            parse_user_input("message", "[code]g")

    def test_no_closing_embedded(self):
        with self.chk_err("BBCode de fermeture [/i] mal placé."):
            parse_user_input("message", "[i]a[b]g[/i]")

    def test_closing_mismatch(self):
        with self.chk_err("BBCode de fermeture [/i] mal placé."):
            parse_user_input("message", "[b]g[/i]")

    def test_no_opening(self):
        with self.chk_err("BBCode de fermeture [/i] mal placé."):
            parse_user_input("message", "[/i]")

    def test_invalid_nesting(self):
        with self.chk_err("Le BBCode [list] ne peut contenir de BBCode [b]."):
            parse_user_input("message", "[list][b]Hello[/b][/list]")

        with self.chk_err("Le BBCode [*] ne peut se trouver que dans un " "BBCode [list]."):
            parse_user_input("message", "[*]Hello[/*]")

    def test_unexpected_parameter(self):
        with self.chk_err("Le BBCode [b] ne prend pas de paramètre."):
            parse_user_input("message", "[b=aaaaa][/b]")

    def test_empty_brackets(self):
        node = parse_user_input("message", "[]...[/]")
        self.chk_bbcode(node, "[]...[/]")
        self.chk_html(node, "[]...[/]")
        self.chk_text(node, "[]...[/]")

    def test_unknown_codes(self):
        node = parse_user_input("message", "[a]...[/x]")
        self.chk_bbcode(node, "[a]...[/x]")
        self.chk_html(node, "[a]...[/x]")
        self.chk_text(node, "[a]...[/x]")

    def test_strike_node(self):
        node = parse_user_input("message", "[s]strike me[/s]")
        self.chk_bbcode(node, "[s]strike me[/s]")
        self.chk_html(node, "<del>strike me</del>")
        self.chk_text(node, "strike me")

    def test_paragraph_node(self):
        node = parse_user_input("web_page", "[p]n[/p][p=left]n[/p][p=justify]n[/p]")
        self.chk_bbcode(node, "[p]n[/p][p=left]n[/p][p=justify]n[/p]")
        self.chk_html(node, '<p>n</p><p class="alignLeft">n</p><p class="alignJustify">n</p>')
        self.chk_text(node, "\nn\n\nn\n\nn\n")

        with self.chk_err("Paramètre d’alignement incorrect"):
            parse_user_input("web_page", "[p=sdfsfds][/p]")

    def test_quote_node(self):
        node = parse_user_input("message", "[quote]qtext[/quote]\nx")
        self.chk_bbcode(node, "[quote]qtext[/quote]\nx")
        self.chk_html(
            node,
            '<div class="quote"><span class="quote">❝</span>' " <br />\nqtext</div>\nx",
        )
        self.chk_text(node, "\n« qtext »\n\nx")

        node = parse_user_input("message", "[quote='\"abcd[]qtext[/quote]")
        self.chk_bbcode(node, "[quote='\"abcd[]qtext[/quote]")
        self.chk_html(
            node,
            '<div class="quote"><span class="quote">❝</span>'
            ' <span class="quoted">&#x27;&quot;abcd[</span><br />\nqtext</div>',
        )
        self.chk_text(node, "\n« qtext »\n")

    def test_code_node(self):
        node = parse_user_input("message", "[code][/codex][b]x[/b][/i][[/code]")
        self.chk_bbcode(node, "[code][/codex][b]x[/b][/i][[/code]")
        self.chk_html(node, "<pre>[/codex][b]x[/b][/i][</pre>")
        self.chk_text(node, "\n[Code]\n")

    def test_list_node(self):
        node = parse_user_input("message", "[list][*]a[/*][*]b[/*][/list]")
        self.chk_bbcode(node, "[list][*]a[/*][*]b[/*][/list]")
        self.chk_html(node, "<ul><li>a</li><li>b</li></ul>")
        self.chk_text(node, "\n    • a\n    • b\n")

        node = parse_user_input("message", "[list][*]a[*]b[/list]")
        self.chk_bbcode(node, "[list][*]a[/*][*]b[/*][/list]")
        self.chk_html(node, "<ul><li>a</li><li>b</li></ul>")
        self.chk_text(node, "\n    • a\n    • b\n")

    def test_url_node(self):
        """Test the [url] BBCode tag"""
        # simple url
        node = parse_user_input("message", "[url]http://www.bde.enseeiht.fr/[/url]")
        self.chk_bbcode(node, "[url]http://www.bde.enseeiht.fr/[/url]")
        self.chk_html(
            node,
            '<a href="http://www.bde.enseeiht.fr/">' "http://www.bde.enseeiht.fr/</a>",
        )
        self.chk_text(node, "http://www.bde.enseeiht.fr/")

        # mailto
        node = parse_user_input("message", "[url]mailto:test@example.org[/url]")
        self.chk_bbcode(node, "[url]mailto:test@example.org[/url]")
        self.chk_html(node, '<a href="mailto:test@example.org">' "mailto:test@example.org</a>")
        self.chk_text(node, "mailto:test@example.org")

        # javascript
        with self.chk_err("URL javascript:blabla invalide."):
            parse_user_input("message", "[url]javascript:blabla[/url]")

        # url with description
        node = parse_user_input("message", "[url=http://perdu.com/&]Perdu ?[/url]")
        self.chk_bbcode(node, "[url=http://perdu.com/&]Perdu ?[/url]")
        self.chk_html(node, '<a href="http://perdu.com/&amp;">Perdu ?</a>')
        self.chk_text(node, "Perdu ?")

        # url with image
        node = parse_user_input(
            "message",
            "[url=http://perdu.com]Perdu ?[img]http://example.org/image.png" "[/img][/url]",
        )
        self.chk_bbcode(
            node,
            "[url=http://perdu.com]Perdu ?[img]" "http://example.org/image.png[/img][/url]",
        )
        self.chk_html(
            node,
            '<a href="http://perdu.com">Perdu ?' '<img src="http://example.org/image.png" alt="" /></a>',
        )
        self.chk_text(node, "Perdu ?")

        # url with bold text
        node = parse_user_input("message", "[url=http://perdu.com]Perdu ?[b]x[/b][/url]")
        self.chk_bbcode(node, "[url=http://perdu.com]Perdu ?[b]x[/b][/url]")
        self.chk_html(node, '<a href="http://perdu.com">Perdu ?<b>x</b></a>')
        self.chk_text(node, "Perdu ?x")

        # URL without parameters and non-text contents
        with self.chk_err("URL  invalide."):
            parse_user_input("message", "[url][b][/b][/url]")

        with self.chk_err("URL  invalide."):
            parse_user_input("message", "[url][b]a[/b][/url]")

        with self.chk_err("URL  invalide."):
            parse_user_input("message", "[url]http://perdu.com[b][/b][/url]")

    def test_img_node(self):
        node = parse_user_input("message", "[img=Une image]http://example.org/image.png[/img]")
        self.chk_bbcode(node, "[img=Une image]http://example.org/image.png[/img]")
        self.chk_html(node, '<img src="http://example.org/image.png" ' 'alt="Une image" />')
        self.chk_text(node, "Une image")

        with self.chk_err("URL d’image invalide."):
            parse_user_input("message", "[img=Une image]bla[/img]")

    def test_size_node(self):
        node = parse_user_input("message", "[size=120]test[/size]")
        self.chk_bbcode(node, "[size=120]test[/size]")
        self.chk_html(node, '<span style="font-size:120%">test</span>')
        self.chk_text(node, "test")

        with self.chk_err("Taille de texte invalide."):
            parse_user_input("message", "[size=sdfdfs]test[/size]")

        with self.chk_err("Taille de texte invalide."):
            parse_user_input("message", "[size=0]test[/size]")

        with self.chk_err("Taille de texte invalide."):
            parse_user_input("message", "[size=-20]test[/size]")

        with self.chk_err("Taille de texte invalide."):
            parse_user_input("message", "[size=201]test[/size]")

    def test_color_node(self):
        node = parse_user_input("message", "[color=#FF00FF]test[/color]")
        self.chk_bbcode(node, "[color=#FF00FF]test[/color]")
        self.chk_html(node, '<span style="color:#FF00FF">test</span>')
        self.chk_text(node, "test")

        with self.chk_err("Couleur invalide."):
            parse_user_input("message", "[color=#F]test[/color]")

        with self.chk_err("Couleur invalide."):
            parse_user_input("message", "[color=#&éàç]test[/color]")

        with self.chk_err("Couleur invalide."):
            parse_user_input("message", "[color=#]test[/color]")

        with self.chk_err("Le BBCode [color] nécessite un paramètre."):
            parse_user_input("message", "[color]test[/color]")

    def test_color_nested(self):
        node = parse_user_input("web_page", "[h3][color=#0000FF]test[/color][/h3]")
        self.chk_bbcode(node, "[h3][color=#0000FF]test[/color][/h3]")
        self.chk_html(node, '<h3><span style="color:#0000FF">test</span></h3>')
        self.chk_text(node, "test")

    def test_heading(self):
        node = parse_user_input("web_page", "[h1]test[/h1]")
        self.chk_bbcode(node, "[h1]test[/h1]")
        self.chk_html(node, "<h1>test</h1>")
        self.chk_text(node, "test")

    def test_attachment(self):
        node = parse_user_input("message_attach", "x: [attach]test. é[/attach]")
        self.chk_bbcode(node, "x: [attach]test. é[/attach]")

        self.assertEqual(node.get_attachments(), ["test. é"])

        with self.assertRaises(AssertionError):
            node.as_html(attachment_callback=None)

        def callback(name):
            return f"<<{name}>>"

        self.chk_html(node, "x: <<test. é>>", attachment_callback=callback)
        self.chk_text(node, "x: ")

    def chk_bbcode(self, node, value):
        self.assertEqual(node.as_bbcode(), value)

    def chk_html(self, node, value, **kwargs):
        self.assertEqual(node.as_html(**kwargs), value)

    def chk_text(self, node, value):
        self.assertEqual(node.as_text(), value)

    def chk_err(self, message):
        message = re.escape(message)

        return self.assertRaisesRegex(BBCodeParseError, f"^{message}$")
